# -*- coding: utf-8 -*-
"""
Utilities for helping to plot cellpy-data.
"""

import collections
import importlib
import itertools
import logging
import os
import sys
import warnings
from io import StringIO
from pathlib import Path

from cellpy.parameters.internal_settings import (
    get_headers_journal,
    get_headers_normal,
    get_headers_step_table,
    get_headers_summary,
)
from cellpy.utils import helpers

try:
    import matplotlib.pyplot as plt

    plt_available = True
except ImportError:
    plt_available = False

try:
    import holoviews as hv
    from holoviews import opts
    from holoviews.plotting.links import RangeToolLink

    hv_available = True
except ImportError:
    hv_available = False

bokeh_available = importlib.util.find_spec("bokeh") is not None

# logger = logging.getLogger(__name__)
logging.captureWarnings(True)

SYMBOL_DICT = {
    "all": [
        "s",
        "o",
        "v",
        "^",
        "<",
        ">",
        "D",
        "p",
        "*",
        "1",
        "2",
        ".",
        ",",
        "3",
        "4",
        "8",
        "p",
        "d",
        "h",
        "H",
        "+",
        "x",
        "X",
        "|",
        "_",
    ],
    "simple": ["s", "o", "v", "^", "<", ">", "*", "d"],
}

COLOR_DICT = {
    "classic": ["b", "g", "r", "c", "m", "y", "k"],
    "grayscale": ["0.00", "0.40", "0.60", "0.70"],
    "bmh": [
        "#348ABD",
        "#A60628",
        "#7A68A6",
        "#467821",
        "#D55E00",
        "#CC79A7",
        "#56B4E9",
        "#009E73",
        "#F0E442",
        "#0072B2",
    ],
    "dark_background": [
        "#8dd3c7",
        "#feffb3",
        "#bfbbd9",
        "#fa8174",
        "#81b1d2",
        "#fdb462",
        "#b3de69",
        "#bc82bd",
        "#ccebc4",
        "#ffed6f",
    ],
    "ggplot": [
        "#E24A33",
        "#348ABD",
        "#988ED5",
        "#777777",
        "#FBC15E",
        "#8EBA42",
        "#FFB5B8",
    ],
    "fivethirtyeight": ["#30a2da", "#fc4f30", "#e5ae38", "#6d904f", "#8b8b8b"],
    "seaborn-colorblind": [
        "#0072B2",
        "#009E73",
        "#D55E00",
        "#CC79A7",
        "#F0E442",
        "#56B4E9",
    ],
    "seaborn-deep": ["#4C72B0", "#55A868", "#C44E52", "#8172B2", "#CCB974", "#64B5CD"],
    "seaborn-bright": [
        "#003FFF",
        "#03ED3A",
        "#E8000B",
        "#8A2BE2",
        "#FFC400",
        "#00D7FF",
    ],
    "seaborn-muted": ["#4878CF", "#6ACC65", "#D65F5F", "#B47CC7", "#C4AD66", "#77BEDB"],
    "seaborn-pastel": [
        "#92C6FF",
        "#97F0AA",
        "#FF9F9A",
        "#D0BBFF",
        "#FFFEA3",
        "#B0E0E6",
    ],
    "seaborn-dark-palette": [
        "#001C7F",
        "#017517",
        "#8C0900",
        "#7600A1",
        "#B8860B",
        "#006374",
    ],
}

hdr_summary = get_headers_summary()
hdr_raw = get_headers_normal()
hdr_steps = get_headers_step_table()
hdr_journal = get_headers_journal()


def _hv_bokeh_available():
    if not hv_available:
        print("You need holoviews. But I cannot load it. Aborting...")
        return False
    if not bokeh_available:
        print("You need Bokeh. But I cannot find it. Aborting...")
        return False
    return True


def find_column(columns, label=None, end="cycle_index"):
    """find columns based on how the column-name ends.

    Args:
        columns: pandas columns
        label: if not provided, generate, if provided, return as is
        end: the string to use for searching

    Returns:
        column header, label
    """
    hdr = None
    lab = None
    for col in columns:
        if col.endswith(end):
            hdr = col
            if label is None:
                lab = col.replace("_", " ")
            else:
                lab = label
            break
    return hdr, lab


def plot_concatenated(
    dataframe,
    x=None,
    y=None,
    err=None,
    xlabel=None,
    ylabel=None,
    points=True,
    line=True,
    errors=True,
    hover=True,
    width=800,
    height=300,
    journal=None,
    file_id_level=0,
    hdr_level=None,
    axis=1,
    mean_end="_mean",
    std_end="_std",
    cycle_end="cycle_index",
    legend_title="cell-type",
    marker_size=None,
    cmap="default_colors",
    spread=False,
    extension="bokeh",
    edges=False,
    keys=None,
    simple=False,
    **kwargs,
):
    """Create a holoviews plot of the concatenated summary.

    This function is still under development. Feel free to contribute.

    Args:
        dataframe: the concatenated summary
        x: colum-name for the x variable (not implemented yet)
        y: colum-name for the y variable (not implemented yet)
        err: colum-name for the std variable (not implemented yet)
        xlabel: label for x-axis
        ylabel: label for y-axis
        points (bool): plot points if True
        line (bool): plot line if True
        errors (bool): plot errors if True
        hover (bool): add hover tool if True
        width: width of plot
        height: height of plot
        journal: batch.journal object
        file_id_level: the level (multiindex-level) where the cell-names are.
        hdr_level:  the level (multiindex-level) where the parameter names are.
        axis: what axis to use when looking in the data-frame (row-based or col-based).
        mean_end: used for searching for y-column names
        std_end: used for searching for e-column names
        cycle_end: used for searching for x-column name
        legend_title: title to put over the legends
        marker_size: size of the markers used
        cmap: color-map to use
        spread (bool): plot error-bands instead of error-bars if True
        extension (str): "matplotlib" or "bokeh". Note, this uses hv.extension) and will affect the
            state of your notebook
        edges (bool): show all axes
        keys (dict): columns to plot (not working yet)
        simple (bool): making a simple hv.Overlay instead of an hv.NdOverlay if True
        **kwargs: key-word arguments sent to hv.NdOverlay

    Example:
        >>> my_mpl_plot = plot_concatenated(
        >>>     cap_cycle_norm_fast_1000, journal=b.experiment.journal,
        >>>     height=500, marker_size=5,
        >>>     extension="matplotlib",
        >>>     edges=True,
        >>> )

        >>> my_bokeh_plot = plot_concatenated(
        >>>     cap_cycle_norm_fast_1000, journal=b.experiment.journal,
        >>>     height=500, marker_size=5,
        >>>     edges=True,
        >>> )


    Example:
        >>> # Simple conversion from bokeh to matplotlib
        >>> # NB! make sure you have only used matplotlib-bokeh convertable key-words (not marker_size)

        >>> hv.extension("matplotlib")
        >>> my_plot.opts(aspect="auto", fig_inches=(12,7), fig_size=90, legend_position="top_right",
        >>>              legend_cols = 2,
        >>>              show_frame=True)

    """
    # TODO: add option for using labels from journal in the legend

    if keys is None:
        keys = dict()

    if not hv_available:
        print(
            "This function uses holoviews. But could not import it."
            "So I am aborting..."
        )
        return

    if extension == "matplotlib":
        hover = False
    elif extension == "plotly":
        print("The plotly backend might not work properly yet.")
        print("Fingers crossed.")
        print(
            "(at least, make sure you are using the most "
            "recent versions of jupyter, holoviews and plotly)"
        )

    try:
        current_extension = hv.Store.current_backend
        if extension != current_extension:
            hv.extension(extension, logo=False)
    except Exception as e:
        hv.extension(extension, logo=False)

    if hdr_level is None:
        hdr_level = 0 if file_id_level == 1 else 1

    averaged = True
    columns = list(set(dataframe.columns.get_level_values(hdr_level)))
    hdr_x, lab_x = find_column(columns, label=xlabel, end=cycle_end)
    hdr_y, lab_y = find_column(columns, label=ylabel, end=mean_end)

    if hdr_y is None:
        averaged = False
        errors = False
        if hdr_x is not None:
            columns.remove(hdr_x)
        hdr_y = columns[0]
        if ylabel is None:
            lab_y = hdr_y.replace("_", " ")
        else:
            lab_y = ylabel
    if errors:
        hdr_e, _ = find_column(columns, end=std_end)

    grouped = dataframe.groupby(axis=axis, level=file_id_level)
    curve_dict = dict()

    if not averaged and journal is not None:
        journal_pages = journal.pages[
            [hdr_journal["group"], hdr_journal["sub_group"]]
        ].copy()
        journal_pages["g"] = 0
        journal_pages["sg"] = 0
        markers = itertools.cycle(["s", "o", "<", "*", "+", "x"])
        colors = itertools.cycle(hv.Cycle(cmap).values)

        j = journal_pages.groupby(hdr_journal["group"])
        for i, (jn, jg) in enumerate(j):
            journal_pages.loc[journal_pages["group"] == jn, "g"] = i
            journal_pages.loc[journal_pages["group"] == jn, "sg"] = list(range(len(jg)))

        markers = [next(markers) for _ in range(journal_pages["sg"].max() + 1)]
        colors = [next(colors) for _ in range(journal_pages["g"].max() + 1)]
        journal_pages = journal_pages[["g", "sg"]]

    for i, (name, group) in enumerate(grouped):
        if name in keys:
            label = keys[name]
        else:
            label = name
            keys[name] = name

        group.columns = group.columns.droplevel(file_id_level)
        if hdr_x is None:
            group = group.reset_index()
            hdr_x = group.columns[0]

        if lab_x is None:
            lab_x = hdr_x.replace("_", " ")

        if not averaged and journal is not None:
            g = journal_pages.loc[name, "g"]
            sg = journal_pages.loc[name, "sg"]
            color = colors[g]
            marker = markers[sg]
            curve = hv.Curve(group, (hdr_x, lab_x), (hdr_y, lab_y), label=label).opts(
                color=color
            )

        else:
            curve = hv.Curve(group, (hdr_x, lab_x), (hdr_y, lab_y), label=label)

        if points:
            if not averaged and journal is not None:

                scatter = hv.Scatter(curve).opts(color=color, marker=marker)

                if edges and extension == "matplotlib":
                    scatter = scatter.opts(edgecolor="k")

                if edges and extension == "bokeh":
                    scatter = scatter.opts(line_color="k", line_width=1)

                if marker_size is not None and extension == "bokeh":
                    scatter = scatter.opts(size=marker_size)
            else:
                scatter = hv.Scatter(curve)

                if marker_size is not None and extension == "bokeh":
                    scatter = scatter.opts(size=marker_size)

        if points and line:
            curve *= scatter

        elif points:
            curve = scatter

        if errors:
            if spread:
                curve *= hv.Spread(group, hdr_x, [hdr_y, hdr_e])
            else:
                curve *= hv.ErrorBars(
                    group, hdr_x, [hdr_y, hdr_e]
                )  # should get the color from curve and set it here
        curve_dict[label] = curve

    if extension == "matplotlib":
        overlay_opts = {
            "aspect": "auto",
            "fig_inches": (width * 0.016, height * 0.012),
            "show_frame": True,
        }
    else:
        overlay_opts = {"width": width, "height": height}

    if simple:
        if len(keys) == len(curve_dict):
            new_curve_dict = {}
            for k in keys:
                new_curve_dict[k] = curve_dict[keys[k]]
            curve_dict = new_curve_dict

        final_plot = hv.Overlay(
            [*curve_dict.values()], vdims=[*curve_dict.keys()]
        ).opts(**overlay_opts, **kwargs)
    else:
        final_plot = hv.NdOverlay(curve_dict, kdims=legend_title).opts(
            **overlay_opts, **kwargs
        )

    if hover and not extension == "plotly":
        if points:
            final_plot.opts(opts.Scatter(tools=["hover"]))
        else:
            final_plot.opts(opts.Curve(tools=["hover"]))
    return final_plot


def create_colormarkerlist_for_journal(
    journal, symbol_label="all", color_style_label="seaborn-colorblind"
):
    """Fetch lists with color names and marker types of correct length for a journal.

    Args:
        journal: cellpy journal
        symbol_label: sub-set of markers to use
        color_style_label: cmap to use for colors

    Returns:
        colors (list), markers (list)
    """
    logging.debug("symbol_label: " + symbol_label)
    logging.debug("color_style_label: " + color_style_label)
    groups = journal.pages[hdr_journal.group].unique()
    sub_groups = journal.pages[hdr_journal.subgroup].unique()
    return create_colormarkerlist(groups, sub_groups, symbol_label, color_style_label)


def create_colormarkerlist(
    groups, sub_groups, symbol_label="all", color_style_label="seaborn-colorblind"
):
    """Fetch lists with color names and marker types of correct length.

    Args:
        groups: list of group numbers (used to generate the list of colors)
        sub_groups: list of sub-group numbers (used to generate the list of markers).
        symbol_label: sub-set of markers to use
        color_style_label: cmap to use for colors

    Returns:
        colors (list), markers (list)
    """
    symbol_list = SYMBOL_DICT[symbol_label]
    color_list = COLOR_DICT[color_style_label]

    # checking that we have enough colors and symbols (if not, then use cycler (e.g. reset))
    color_cycler = itertools.cycle(color_list)
    symbol_cycler = itertools.cycle(symbol_list)
    _color_list = []
    _symbol_list = []
    for i in groups:
        _color_list.append(next(color_cycler))
    for i in sub_groups:
        _symbol_list.append(next(symbol_cycler))
    return _color_list, _symbol_list


def _raw_plot(raw_curve, title="Voltage versus time", **kwargs):

    tgt = raw_curve.relabel(title).opts(
        width=800,
        height=300,
        labelled=["y"],
        # tools=["pan","box_zoom", "reset"],
        active_tools=["pan"],
    )
    src = raw_curve.opts(width=800, height=100, yaxis=None, default_tools=[])

    RangeToolLink(src, tgt)

    layout = (tgt + src).cols(1)
    layout.opts(opts.Layout(shared_axes=False, merge_tools=False))
    return layout


def raw_plot(cell, y=("voltage", "Voltage (V vs Li/Li+)"), title=None, **kwargs):
    # TODO: missing doc-string

    if title is None:
        if isinstance(y, (list, tuple)):
            pre_title = str(y[0])
        else:
            pre_title = str(y)
        title = " ".join([pre_title, "versus", "time"])

    if not _hv_bokeh_available():
        return

    hv.extension("bokeh", logo=False)

    raw = cell.cell.raw
    raw["test_time_hrs"] = raw[hdr_raw["test_time_txt"]] / 3600
    x = ("test_time_hrs", "Time (hours)")
    raw_curve = hv.Curve(raw, x, y)
    layout = _raw_plot(raw_curve, title=title, **kwargs)
    return layout


def cycle_info_plot(
    cell,
    cycle=None,
    step=None,
    title=None,
    points=False,
    x=None,
    y=None,
    info_level=1,
    h_cycle=None,
    h_step=None,
    show_it=True,
    label_cycles=True,
    label_steps=False,
    get_axes=False,
    use_bokeh=True,
    **kwargs,
):
    """Show raw data together with step and cycle information.

    Args:
        cell: cellpy object
        cycle: cycles to select (optional, default is all)
        step: steps to select (optional, defaults is all)
        title: a title to give the plot
        points: overlay a scatter plot
        x (str): column header for the x-value (defaults to "Test_Time")
        y (str): column header for the y-value (defaults to "Voltage")
        info_level (int): how much information to display (defaults to 1)
            (0 - almost nothing
            1 - pretty much
            2 - something else
            3 - not implemented yet).
        h_cycle: column header for the cycle number (defaults to "Cycle_Index")
        h_step: column header for the step number (defaults to "Step_Index")
        show_it (bool): show the figure (defaults to True). If not, return the figure.
        label_cycles (bool): add labels with cycle numbers.
        label_steps (bool): add labels with step numbers
        get_axes (bool): return axes (for matplotlib)
        use_bokeh (bool): use bokeh to plot (defaults to True). If not, use matplotlib.
        **kwargs: parameters specific to either matplotlib or bokeh.

    Returns:
        ``matplotlib.axes`` or None
    """
    # TODO: missing doc-string
    if use_bokeh and not bokeh_available:
        print("OBS! bokeh is not available - using matplotlib instead")
        use_bokeh = False

    if use_bokeh:
        axes = _cycle_info_plot_bokeh(
            cell,
            cycle=cycle,
            step=step,
            title=title,
            points=points,
            x=x,
            y=y,
            info_level=info_level,
            h_cycle=h_cycle,
            h_step=h_step,
            show_it=show_it,
            label_cycles=label_cycles,
            label_steps=label_steps,
            **kwargs,
        )
    else:
        if isinstance(cycle, (list, tuple)):
            if len(cycle) > 1:
                print("OBS! The matplotlib-plotter only accepts single cycles.")
                print(f"Selecting first cycle ({cycle[0]})")
            cycle = cycle[0]
        axes = _cycle_info_plot_matplotlib(cell, cycle, get_axes)
    if get_axes:
        return axes


def _plot_step(ax, x, y, color):
    ax.plot(x, y, color=color, linewidth=3)


def _get_info(table, cycle, step):
    # obs! hard-coded col-names. Please fix me.
    m_table = (table.cycle == cycle) & (table.step == step)
    p1, p2 = table.loc[m_table, ["point_min", "point_max"]].values[0]
    c1, c2 = table.loc[m_table, ["current_min", "current_max"]].abs().values[0]
    d_voltage, d_current = table.loc[
        m_table, ["voltage_delta", "current_delta"]
    ].values[0]
    d_discharge, d_charge = table.loc[
        m_table, ["discharge_delta", "charge_delta"]
    ].values[0]
    current_max = (c1 + c2) / 2
    rate = table.loc[m_table, "rate_avr"].values[0]
    step_type = table.loc[m_table, "type"].values[0]
    return [step_type, rate, current_max, d_voltage, d_current, d_discharge, d_charge]


def _add_step_info_cols(df, table, cycles=None, steps=None, h_cycle=None, h_step=None):
    if h_cycle is None:
        h_cycle = "cycle_index"  # edit
    if h_step is None:
        h_step = "step_index"  # edit

    col_name_mapper = {"cycle": h_cycle, "step": h_step}

    df = df.merge(
        table.rename(columns=col_name_mapper),
        on=("cycle_index", "step_index"),
        how="left",
    )

    return df


def _cycle_info_plot_bokeh(
    cell,
    cycle=None,
    step=None,
    title=None,
    points=False,
    x=None,
    y=None,
    info_level=0,
    h_cycle=None,
    h_step=None,
    show_it=False,
    label_cycles=True,
    label_steps=False,
    **kwargs,
):
    """Plot raw data with annotations.

    This function uses Bokeh for plotting and is intended for use in
    Jupyter Notebooks.
    """
    # TODO: check that correct new column-names are used
    # TODO: fix bokeh import (use e.g. import bokeh.io)

    try:
        from bokeh.io import output_notebook, show
        from bokeh.layouts import column, row
        from bokeh.models import ColumnDataSource, HoverTool, LabelSet
        from bokeh.models.annotations import Span
        from bokeh.models.widgets import Slider, TextInput
        from bokeh.plotting import figure

    except ImportError:
        warnings.warn("Could not import bokeh")
        return

    try:
        output_notebook(hide_banner=True)
    finally:
        sys.stdout = sys.__stdout__

    if points:
        if cycle is None or (len(cycle) > 1):
            print("Plotting points only allowed when plotting one single cycle.")
            print("Turning points off.")
            points = False

    if h_cycle is None:
        h_cycle = "cycle_index"  # edit
    if h_step is None:
        h_step = "step_index"  # edit

    if x is None:
        x = "test_time"  # edit
    if y is None:
        y = "voltage"  # edit

    if isinstance(x, tuple):
        x, x_label = x
    else:
        x_label = x

    if isinstance(y, tuple):
        y, y_label = y
    else:
        y_label = y

    t_x = x  # used in generating title - replace with a selector
    t_y = y  # used in generating title - replace with a selector

    if title is None:
        title = f"{t_y} vs. {t_x}"

    cols = [x, y]
    cols.extend([h_cycle, h_step])

    df = cell.cell.raw.loc[:, cols]

    if cycle is not None:
        if not isinstance(cycle, (list, tuple)):
            cycle = [cycle]

        _df = df.loc[df[h_cycle].isin(cycle), :]
        if len(cycle) < 5:
            title += f" [c:{cycle}]"
        else:
            title += f" [c:{cycle[0]}..{cycle[-1]}]"
        if _df.empty:
            print(f"EMPTY (available cycles: {df[h_step].unique()})")
            return
        else:
            df = _df

    cycle = df[h_cycle].unique()

    if step is not None:
        if not isinstance(step, (list, tuple)):
            step = [step]

        _df = df.loc[df[h_step].isin(step), :]
        if len(step) < 5:
            title += f" (s:{step})"
        else:
            title += f" [s:{step[0]}..{step[-1]}]"
        if _df.empty:
            print(f"EMPTY (available steps: {df[h_step].unique()})")
            return
        else:
            df = _df

    x_min, x_max = df[x].min(), df[x].max()
    y_min, y_max = df[y].min(), df[y].max()

    if info_level > 0:
        table = cell.cell.steps
        df = _add_step_info_cols(df, table, cycle, step)

    source = ColumnDataSource(df)

    plot = figure(
        title=title,
        tools="pan,reset,save,wheel_zoom,box_zoom,undo,redo",
        x_range=[x_min, x_max],
        y_range=[y_min, y_max],
        **kwargs,
    )

    plot.line(x, y, source=source, line_width=3, line_alpha=0.6)

    # labelling cycles
    if label_cycles:
        cycle_line_positions = [df.loc[df[h_cycle] == c, x].min() for c in cycle]
        cycle_line_positions.append(df.loc[df[h_cycle] == cycle[-1], x].max())
        for m in cycle_line_positions:
            _s = Span(
                location=m,
                dimension="height",
                line_color="red",
                line_width=3,
                line_alpha=0.5,
            )
            plot.add_layout(_s)

        s_y_pos = y_min + 0.9 * (y_max - y_min)
        s_x = []
        s_y = []
        s_l = []

        for s in cycle:
            s_x_min = df.loc[df[h_cycle] == s, x].min()
            s_x_max = df.loc[df[h_cycle] == s, x].max()
            s_x_pos = (s_x_min + s_x_max) / 2
            s_x.append(s_x_pos)
            s_y.append(s_y_pos)
            s_l.append(f"c{s}")

        c_labels = ColumnDataSource(data={x: s_x, y: s_y, "names": s_l})

        c_labels = LabelSet(
            x=x,
            y=y,
            text="names",
            level="glyph",
            source=c_labels,
            render_mode="canvas",
            text_color="red",
            text_alpha=0.7,
        )

        plot.add_layout(c_labels)

        # labelling steps
    if label_steps:
        for c in cycle:
            step = df.loc[df[h_cycle] == c, h_step].unique()
            step_line_positions = [
                df.loc[(df[h_step] == s) & (df[h_cycle] == c), x].min()
                for s in step[0:]
            ]
            for m in step_line_positions:
                _s = Span(
                    location=m,
                    dimension="height",
                    line_color="olive",
                    line_width=3,
                    line_alpha=0.1,
                )
                plot.add_layout(_s)

            # s_y_pos = y_min + 0.8 * (y_max - y_min)
            s_x = []
            s_y = []
            s_l = []

            for s in step:
                s_x_min = df.loc[(df[h_step] == s) & (df[h_cycle] == c), x].min()
                s_x_max = df.loc[(df[h_step] == s) & (df[h_cycle] == c), x].max()
                s_x_pos = s_x_min

                s_y_min = df.loc[(df[h_step] == s) & (df[h_cycle] == c), y].min()
                s_y_max = df.loc[(df[h_step] == s) & (df[h_cycle] == c), y].max()
                s_y_pos = (s_y_max + s_y_min) / 2

                s_x.append(s_x_pos)
                s_y.append(s_y_pos)
                s_l.append(f"s{s}")

            s_labels = ColumnDataSource(data={x: s_x, y: s_y, "names": s_l})

            s_labels = LabelSet(
                x=x,
                y=y,
                text="names",
                level="glyph",
                source=s_labels,
                render_mode="canvas",
                text_color="olive",
                text_alpha=0.3,
            )

            plot.add_layout(s_labels)

    hover = HoverTool()
    if info_level == 0:
        hover.tooltips = [
            (x, "$x{0.2f}"),
            (y, "$y"),
            ("cycle", f"@{h_cycle}"),
            ("step", f"@{h_step}"),
        ]
    elif info_level == 1:
        # insert C-rates etc here
        hover.tooltips = [
            (f"(x,y)", "($x{0.2f} $y"),
            ("cycle", f"@{h_cycle}"),
            ("step", f"@{h_step}"),
            ("step_type", "@type"),
            ("rate", "@rate_avr{0.2f}"),
        ]

    elif info_level == 2:
        hover.tooltips = [
            (x, "$x{0.2f}"),
            (y, "$y"),
            ("cycle", f"@{h_cycle}"),
            ("step", f"@{h_step}"),
            ("step_type", "@type"),
            ("rate (C)", "@rate_avr{0.2f}"),
            ("dv (%)", "@voltage_delta{0.2f}"),
            ("I-max (A)", "@current_max"),
            ("I-min (A)", "@current_min"),
            ("dCharge (%)", "@charge_delta{0.2f}"),
            ("dDischarge (%)", "@discharge_delta{0.2f}"),
        ]

    hover.mode = "vline"
    plot.add_tools(hover)

    plot.xaxis.axis_label = x_label
    plot.yaxis.axis_label = y_label

    if points:
        plot.scatter(x, y, source=source, alpha=0.3)

    if show_it:
        show(plot)

    return plot


def _cycle_info_plot_matplotlib(cell, cycle, get_axes=False):

    # obs! hard-coded col-names. Please fix me.
    if not plt_available:
        print(
            "This function uses matplotlib. But I could not import it. "
            "So I decided to abort..."
        )
        return

    data = cell.cell.raw
    table = cell.cell.steps

    span_colors = ["#4682B4", "#FFA07A"]

    voltage_color = "#008B8B"
    current_color = "#CD5C5C"

    m_cycle_data = data.cycle_index == cycle
    all_steps = data[m_cycle_data]["step_index"].unique()

    color = itertools.cycle(span_colors)

    fig = plt.figure(figsize=(20, 8))
    fig.suptitle(f"Cycle: {cycle}")

    ax3 = plt.subplot2grid((8, 3), (0, 0), colspan=3, rowspan=1, fig=fig)  # steps
    ax4 = plt.subplot2grid((8, 3), (1, 0), colspan=3, rowspan=2, fig=fig)  # rate
    ax1 = plt.subplot2grid((8, 3), (3, 0), colspan=3, rowspan=5, fig=fig)  # data

    ax2 = ax1.twinx()
    ax1.set_xlabel("time (minutes)")
    ax1.set_ylabel("voltage (V vs. Li/Li+)", color=voltage_color)
    ax2.set_ylabel("current (mA)", color=current_color)

    annotations_1 = []  # step number (IR)
    annotations_2 = []  # step number
    annotations_4 = []  # rate

    for i, s in enumerate(all_steps):
        m = m_cycle_data & (data.step_index == s)
        c = data.loc[m, "current"] * 1000
        v = data.loc[m, "voltage"]
        t = data.loc[m, "test_time"] / 60
        step_type, rate, current_max, dv, dc, d_discharge, d_charge = _get_info(
            table, cycle, s
        )
        if len(t) > 1:
            fcolor = next(color)

            info_txt = (
                f"{step_type}\nc-rate = {rate}\ni = |{1000 * current_max:0.2f}| mA\n"
            )
            info_txt += f"delta V = {dv:0.2f} %\ndelta i = {dc:0.2f} %\n"
            info_txt += f"delta C = {d_charge:0.2} %\ndelta DC = {d_discharge:0.2} %\n"

            for ax in [ax2, ax3, ax4]:
                ax.axvspan(t.iloc[0], t.iloc[-1], facecolor=fcolor, alpha=0.2)
            _plot_step(ax1, t, v, voltage_color)
            _plot_step(ax2, t, c, current_color)
            annotations_1.append([f"{s}", t.mean()])
            annotations_4.append([info_txt, t.mean()])
        else:
            info_txt = f"{s}({step_type})"
            annotations_2.append([info_txt, t.mean()])
    ax3.set_ylim(0, 1)
    for s in annotations_1:
        ax3.annotate(f"{s[0]}", (s[1], 0.2), ha="center")

    for s in annotations_2:
        ax3.annotate(f"{s[0]}", (s[1], 0.6), ha="center")

    for s in annotations_4:
        ax4.annotate(f"{s[0]}", (s[1], 0.0), ha="center")

    for ax in [ax3, ax4]:
        ax.axes.get_yaxis().set_visible(False)
        ax.axes.get_xaxis().set_visible(False)

    if get_axes:
        return ax1, ax2, ax2, ax4


def save_fig(figure, file_name=None, wide=False, size=None, dpi=300, **kwargs):
    """Save a figure, either a holoviews plot or a matplotlib figure.

    This function should mainly be used when using the standard cellpy notebook
    template (generated by '> cellpy new')

    Args:
        figure (obj): the figure or plot object
        file_name (str): the file name
        wide (bool): release the equal aspect lock (default on for holoviews)
        size (int or tuple of ints): figure size in inches
        dpi (int): resolution
        **kwargs: sent to cellpy.utils.plotutils.hv_bokeh_to_mpl
    """

    out_path = Path("out/")
    extension = "png"

    if size is None:
        size = (6, 6)

    # create file name:
    if file_name is None:
        counter = 1
        while True:
            _file_name = f"cellpy-plot-{str(counter).zfill(3)}.{extension}"
            _file_name = out_path / _file_name
            if not os.path.isfile(_file_name):
                break
            counter += 1
        file_name = _file_name

    type_str = str(type(figure))

    is_hv = type_str.find("holoviews") >= 0
    is_mpl = type_str.find("matplotlib") >= 0

    if is_mpl:
        is_mpl_figure = type_str.find("figure.Figure") >= 0
        is_mpl_axes = type_str.find(".axes.") >= 0
        if not is_mpl_figure:
            if is_mpl_axes:
                figure = figure.get_figure()
            else:
                print("this matplotlib object is not supported")
                print(type_str)
                return

        figure.savefig(file_name, dpi=dpi)
    elif is_hv:
        is_hv_nd_overlay = isinstance(figure, hv.core.overlay.NdOverlay)
        is_hv_overlay = isinstance(figure, hv.core.overlay.Overlay)
        is_hv_layout = isinstance(figure, hv.core.layout.Layout)

        figure = hv_bokeh_to_mpl(figure, wide=wide, size=size, **kwargs)
        figure.savefig(file_name, dpi=dpi)

    else:
        print("this figure object is not supported")
        print(type_str)
        return
    print(f"saved to {file_name}")


def hv_bokeh_to_mpl(figure, wide=False, size=(6, 4), **kwargs):
    # I think this needs to be tackled differently. For example by setting hv.extension("matplotlib") and
    # re-making the figure. Or making a custom renderer.
    figure = hv.render(figure, backend="matplotlib")
    axes = figure.axes
    number_of_axes = len(axes)
    if number_of_axes > 1:
        for j, ax in enumerate(axes):
            if j < number_of_axes - 1:
                ax.set_xlabel("")
            if j < number_of_axes - 1:
                ax.get_legend().remove()
            else:
                handles, labels = ax.get_legend_handles_labels()
                # This does not work:
                # ax.legend(handles, labels, bbox_to_anchor=(1.05, 1), loc="upper_right")
                # TODO: create new legend based on the ax data

    if wide:
        for axis in axes:
            axis.set_aspect("auto")

    figure.tight_layout()
    figure.set_size_inches(size)

    return figure


def oplot(
    b,
    cap_ylim=None,
    ce_ylim=None,
    ir_ylim=None,
    simple=False,
    group_it=False,
    spread=True,
    capacity_unit="gravimetric",
    **kwargs,
):
    """create a holoviews-plot containing Coulombic Efficiency, Capacity, and IR.

    Args:
        b (cellpy.batch object): the batch with the cells.
        cap_ylim (tuple of two floats): scaling of y-axis for capacity plots.
        ce_ylim (tuple of two floats): scaling of y-axis for c.e. plots.
        ir_ylim (tuple of two floats): scaling of y-axis for i.r. plots.
        simple (bool): if True, use hv.Overlay instead of hv.NdOverlay.
        group_it (bool): if True, average pr group.
        spread (bool): if True, show spread instead of error-bars
        capacity_unit (str): select "gravimetric", or "areal"

    Returns:
        ``hv.Overlay`` or ``hv.NdOverlay``
    """

    extension = kwargs.pop("extension", "bokeh")

    cap_colum_dict = {
        "gravimetric": {
            "discharge": "discharge_capacity",
            "charge": "charge_capacity",
            "unit": "mAh/g(a.m.)",
            "ylim": (0, 5000),
        },
        "areal": {
            "discharge": "areal_discharge_capacity",
            "charge": "areal_charge_capacity",
            "unit": "mAh/cm2",
            "ylim": (0, 3),
        },
    }

    if cap_ylim is not None:
        cap_colum_dict[capacity_unit]["ylim"] = cap_ylim

    if ce_ylim is None:
        ce_ylim = (80, 120)

    if ir_ylim is None:
        ir_ylim = (0, 200)

    overlay_sensitive_opts = {"ce": {}, "dcap": {}, "ccap": {}, "ird": {}, "irc": {}}

    layout_sensitive_opts = {"ce": {}, "dcap": {}, "ccap": {}, "ird": {}, "irc": {}}

    if extension == "bokeh":
        overlay_sensitive_opts["ce"] = {"height": 150}
        overlay_sensitive_opts["dcap"] = {"height": 400}
        overlay_sensitive_opts["ccap"] = {"height": 150}
        overlay_sensitive_opts["ird"] = {"height": 150}
        overlay_sensitive_opts["irc"] = {"height": 150}

    elif extension == "matplotlib":
        simple = True
        overlay_sensitive_opts["ce"] = {"aspect": 2}
        overlay_sensitive_opts["dcap"] = {"aspect": 2}
        overlay_sensitive_opts["ccap"] = {"aspect": 2}
        overlay_sensitive_opts["ird"] = {"aspect": 2}
        overlay_sensitive_opts["irc"] = {"aspect": 2}

        hspace = 2
        layout_sensitive_opts["ce"] = {"hspace": hspace}
        layout_sensitive_opts["dcap"] = {"hspace": hspace}
        layout_sensitive_opts["ccap"] = {"hspace": hspace}
        layout_sensitive_opts["ird"] = {"hspace": hspace}
        layout_sensitive_opts["irc"] = {"hspace": hspace}

    bplot_shared_opts = {
        "group_it": group_it,
        "simple": simple,
        "spread": spread,
        "extension": extension,
    }

    if simple:
        overlay_opts = hv.opts.Overlay
        layout_opts = hv.opts.Layout
    else:
        overlay_opts = hv.opts.NdOverlay
        layout_opts = hv.opts.NdLayout

    # print("creating interactive plots")
    oplot_ce = bplot(
        b, columns=["coulombic_efficiency"], **bplot_shared_opts, **kwargs
    ).opts(
        hv.opts.Curve(ylim=ce_ylim),
        overlay_opts(
            title="",
            show_legend=False,
            xlabel="",
            ylabel="C.E.",
            **overlay_sensitive_opts["ce"],
        ),
        layout_opts(title="Coulombic efficiency (%)", **layout_sensitive_opts["ce"]),
    )
    # print(" - created oplot_ce")

    oplot_dcap = bplot(
        b,
        columns=[cap_colum_dict[capacity_unit]["discharge"]],
        **bplot_shared_opts,
        **kwargs,
    ).opts(
        hv.opts.Curve(ylim=cap_colum_dict[capacity_unit]["ylim"]),
        overlay_opts(
            title="",
            show_legend=True,
            xlabel="",
            ylabel="discharge",
            **overlay_sensitive_opts["dcap"],
        ),
        layout_opts(
            title=f"Capacity ({cap_colum_dict[capacity_unit]['unit']})",
            **layout_sensitive_opts["dcap"],
        ),
    )
    # print(" - created oplot_dcap")

    oplot_ccap = bplot(
        b,
        columns=[cap_colum_dict[capacity_unit]["charge"]],
        **bplot_shared_opts,
        **kwargs,
    ).opts(
        hv.opts.Curve(ylim=cap_colum_dict[capacity_unit]["ylim"]),
        overlay_opts(
            title="",
            show_legend=False,
            xlabel="",
            ylabel="charge",
            **overlay_sensitive_opts["ccap"],
        ),
        layout_opts(title="", **layout_sensitive_opts["ccap"]),
    )
    # print(" - created oplot_ccap")

    oplot_ird = bplot(b, columns=["ir_discharge"], **bplot_shared_opts, **kwargs).opts(
        hv.opts.Curve(ylim=ir_ylim),
        overlay_opts(
            title="",
            show_legend=False,
            xlabel="",
            ylabel="discharge",
            **overlay_sensitive_opts["ird"],
        ),
        layout_opts(title="Internal Resistance (Ohm)", **layout_sensitive_opts["ird"]),
    )

    oplot_irc = bplot(b, columns=["ir_charge"], **bplot_shared_opts, **kwargs).opts(
        hv.opts.Curve(ylim=ir_ylim),
        overlay_opts(
            title="",
            show_legend=False,
            ylabel="charge",
            **overlay_sensitive_opts["irc"],
        ),
        layout_opts(title="", **layout_sensitive_opts["irc"]),
    )

    return (oplot_ce + oplot_dcap + oplot_ccap + oplot_ird + oplot_irc).cols(1)


def bplot(b, individual=False, cols=1, **kwargs):
    """plot batch summaries.

    This is wrapper around the two functions concatenate_summaries and plot_concatenated.

    >>> p1 = bplot(b, columns=["charge_capacity"], journal=b.experiment.journal, group_it=True)

    is equivalent to:

    >>> cs = helpers.concatenate_summaries(b, columns=["charge_capacity"], group_it=True)
    >>> p1 = plot_concatenated(cs, journal=journal)

    Args:
        b (cellpy.batch object): the batch with the cells.
        individual (bool): in case of multiple columns, return a list of plots instaed of a hv.Layout
        cols (int): number of columns.

    Key-word arguments sent further to the concatenator:

    Keyword Args:
        rate (float): filter on rate (C-rate)
        on (str or list of str): only select cycles if based on the rate of this step-type (e.g. on="charge").
        columns (list): selected column(s) (using cellpy name) [defaults to "charge_capacity"]
        column_names (list): selected column(s) (using exact column name)
        normalize_capacity_on (list): list of cycle numbers that will be used for setting the basis of the normalization
            (typically the first few cycles after formation)
        scale_by (float or str): scale the normalized data with nominal capacity if "nom_cap", or given value (defaults to one).
        nom_cap (float): nominal capacity of the cell
        normalize_cycles (bool): perform a normalisation of the cycle numbers (also called equivalent cycle index)
        add_areal (bool):  add areal capacity to the summary
        group_it (bool): if True, average pr group.
        rate_std (float): allow for this inaccuracy when selecting cycles based on rate
        rate_column (str): name of the column containing the C-rates.
        inverse (bool): select steps that do not have the given C-rate.
        inverted (bool): select cycles that do not have the steps filtered by given C-rate.
        journal (batch.journal object): the journal (will use the journal in b if not given).

    Key-word arguments sent further to the plotter:

    Keyword Args:
        width (int): width of plot.
        spread (bool): use error-spread instead of error-bars.
        simple (bool): use ``hv.Overlay`` instead of ``hv.NdOverlay``.

    Returns:
        ``holoviews`` plot
    """
    width = kwargs.pop("width", 800)
    journal = kwargs.pop("journal", b.experiment.journal)
    spread = kwargs.pop("spread", True)
    simple = kwargs.pop("simple", False)
    columns = kwargs.pop("columns", ["charge_capacity"])
    extension = kwargs.pop("extension", "bokeh")
    p = collections.OrderedDict()
    i_width = width // cols
    for i, col in enumerate(columns):
        try:
            cs = helpers.concatenate_summaries(b, columns=[col], **kwargs)
            _p = plot_concatenated(
                cs,
                journal=journal,
                spread=spread,
                width=i_width,
                extension=extension,
                title=col,
                simple=simple,
            )

            if i < len(columns) - 1:
                _p.opts(show_legend=False)
                if cols == 1:
                    _p.opts(xlabel="")
            else:
                _p.opts(show_legend=True, legend_position="right")
            # if (len(columns) > 1) and cols > 1:
            #     _p.opts(frame_width=width)
            if extension == "bokeh":
                _p.opts(frame_width=width)
            p[col] = _p

        except KeyError as e:
            print(f"Sorry - missing key: {col}")
            logging.debug(e)

    w = width / 180 * cols
    h = 5 * len(p) / cols

    if len(p) >= 1:
        if not individual:
            if simple:
                out = hv.Layout(list(p.values())).cols(cols)
            else:
                out = hv.NdLayout(p, sort=False).cols(cols)
            if extension == "matplotlib":
                out.opts(fig_inches=(w, h))
        else:
            if extension == "matplotlib":
                out = [o.opts(fig_inches=(w, h)) for o in p.values()]
            else:
                out = [p.values()]
        return out


if __name__ == "__main__":
    pass



"""Internal legacy definitions for loading old file formats."""

# Supported versions: 3, 4, 5

import logging

from cellpy.parameters.internal_settings import (
    get_headers_normal,
    get_headers_step_table,
    get_headers_summary,
)

HEADERS_NORMAL = get_headers_normal()
HEADERS_SUMMARY = get_headers_summary()
HEADERS_STEP_TABLE = get_headers_step_table()

HEADERS_KEYS_STEP_TABLE_EXTENDED = [
    "point",
    "test_time",
    "step_time",
    "current",
    "voltage",
    "charge",
    "discharge",
    "internal_resistance",
]
HEADERS_STEP_TABLE_EXTENSIONS = ["min", "max", "avr", "first", "last", "delta", "std"]

headers_normal_v5 = dict()
headers_summary_v5 = dict()
headers_step_table_v5 = dict()
headers_journal_v0 = dict()

headers_summary_v5["cycle_index"] = "Cycle_Index"
headers_summary_v5["data_point"] = "Data_Point"
headers_summary_v5["test_time"] = "Test_Time"
headers_summary_v5["datetime"] = "DateTime"
headers_summary_v5["discharge_capacity"] = "Discharge_Capacity(mAh/g)"
headers_summary_v5["charge_capacity"] = "Charge_Capacity(mAh/g)"
headers_summary_v5["discharge_capacity_raw"] = "Discharge_Capacity"
headers_summary_v5["charge_capacity_raw"] = "Charge_Capacity"

headers_summary_v5["cumulated_charge_capacity"] = "Cumulated_Charge_Capacity(mAh/g)"
headers_summary_v5[
    "cumulated_discharge_capacity"
] = "Cumulated_Discharge_Capacity(mAh/g)"
headers_summary_v5["coulombic_efficiency"] = "Coulombic_Efficiency(percentage)"
headers_summary_v5[
    "cumulated_coulombic_efficiency"
] = "Cumulated_Coulombic_Efficiency(percentage)"
headers_summary_v5["coulombic_difference"] = "Coulombic_Difference(mAh/g)"
headers_summary_v5[
    "cumulated_coulombic_difference"
] = "Cumulated_Coulombic_Difference(mAh/g)"
headers_summary_v5["discharge_capacity_loss"] = "Discharge_Capacity_Loss(mAh/g)"
headers_summary_v5["charge_capacity_loss"] = "Charge_Capacity_Loss(mAh/g)"
headers_summary_v5[
    "cumulated_discharge_capacity_loss"
] = "Cumulated_Discharge_Capacity_Loss(mAh/g)"
headers_summary_v5[
    "cumulated_charge_capacity_loss"
] = "Cumulated_Charge_Capacity_Loss(mAh/g)"
headers_summary_v5["ir_discharge"] = "IR_Discharge(Ohms)"
headers_summary_v5["ir_charge"] = "IR_Charge(Ohms)"
headers_summary_v5["ocv_first_min"] = "OCV_First_Min(V)"
headers_summary_v5["ocv_second_min"] = "OCV_Second_Min(V)"
headers_summary_v5["ocv_first_max"] = "OCV_First_Max(V)"
headers_summary_v5["ocv_second_max"] = "OCV_Second_Max(V)"
headers_summary_v5["end_voltage_discharge"] = "End_Voltage_Discharge(V)"
headers_summary_v5["end_voltage_charge"] = "End_Voltage_Charge(V)"
headers_summary_v5["cumulated_ric_disconnect"] = "RIC_Disconnect(none)"
headers_summary_v5["cumulated_ric_sei"] = "RIC_SEI(none)"
headers_summary_v5["cumulated_ric"] = "RIC(none)"

headers_summary_v5["normalized_cycle_index"] = "Normalized_Cycle_Index"
# Sum of irreversible capacity:
headers_summary_v5["low_level"] = "Low_Level(percentage)"
# SEI loss:
headers_summary_v5["high_level"] = "High_Level(percentage)"
headers_summary_v5["shifted_charge_capacity"] = "Charge_Endpoint_Slippage(mAh/g)"
headers_summary_v5["shifted_discharge_capacity"] = "Discharge_Endpoint_Slippage(mAh/g)"
headers_summary_v5["temperature_last"] = "Last_Temperature(C)"
headers_summary_v5["temperature_mean"] = "Average_Temperature(C)"
headers_summary_v5["charge_c_rate"] = "Charge_C_rate"
headers_summary_v5["discharge_c_rate"] = "Discharge_C_rate"
headers_summary_v5["pre_aux"] = "Aux_"

# - normal (data) -

headers_normal_v5["aci_phase_angle_txt"] = "ACI_Phase_Angle"
headers_normal_v5["ref_aci_phase_angle_txt"] = "Reference_ACI_Phase_Angle"

headers_normal_v5["ac_impedance_txt"] = "AC_Impedance"
headers_normal_v5["ref_ac_impedance_txt"] = "Reference_AC_Impedance"  # new

headers_normal_v5["charge_capacity_txt"] = "Charge_Capacity"
headers_normal_v5["charge_energy_txt"] = "Charge_Energy"
headers_normal_v5["current_txt"] = "Current"
headers_normal_v5["cycle_index_txt"] = "Cycle_Index"
headers_normal_v5["data_point_txt"] = "Data_Point"
headers_normal_v5["datetime_txt"] = "DateTime"
headers_normal_v5["discharge_capacity_txt"] = "Discharge_Capacity"
headers_normal_v5["discharge_energy_txt"] = "Discharge_Energy"
headers_normal_v5["internal_resistance_txt"] = "Internal_Resistance"

headers_normal_v5["is_fc_data_txt"] = "Is_FC_Data"
headers_normal_v5["step_index_txt"] = "Step_Index"
headers_normal_v5["sub_step_index_txt"] = "Sub_Step_Index"  # new

headers_normal_v5["step_time_txt"] = "Step_Time"
headers_normal_v5["sub_step_time_txt"] = "Sub_Step_Time"  # new

headers_normal_v5["test_id_txt"] = "Test_ID"
headers_normal_v5["test_time_txt"] = "Test_Time"

headers_normal_v5["voltage_txt"] = "Voltage"
headers_normal_v5["ref_voltage_txt"] = "Reference_Voltage"  # new

headers_normal_v5["dv_dt_txt"] = "dV/dt"
headers_normal_v5["frequency_txt"] = "Frequency"  # new
headers_normal_v5["amplitude_txt"] = "Amplitude"  # new

# - step table -

headers_step_table_v5["test"] = "test"
headers_step_table_v5["ustep"] = "ustep"
headers_step_table_v5["cycle"] = "cycle"
headers_step_table_v5["step"] = "step"
headers_step_table_v5["test_time"] = "test_time"
headers_step_table_v5["step_time"] = "step_time"
headers_step_table_v5["sub_step"] = "sub_step"
headers_step_table_v5["type"] = "type"
headers_step_table_v5["sub_type"] = "sub_type"
headers_step_table_v5["info"] = "info"
headers_step_table_v5["voltage"] = "voltage"
headers_step_table_v5["current"] = "current"
headers_step_table_v5["charge"] = "charge"
headers_step_table_v5["discharge"] = "discharge"
headers_step_table_v5["point"] = "point"
headers_step_table_v5["internal_resistance"] = "ir"
headers_step_table_v5["internal_resistance_change"] = "ir_pct_change"
headers_step_table_v5["rate_avr"] = "rate_avr"


headers_journal_v0["filename"] = "filenames"
headers_journal_v0["mass"] = "masses"
headers_journal_v0["total_mass"] = "total_masses"
headers_journal_v0["loading"] = "loadings"
headers_journal_v0["fixed"] = "fixed"
headers_journal_v0["label"] = "labels"
headers_journal_v0["cell_type"] = "cell_types"
headers_journal_v0["raw_file_names"] = "raw_file_names"
headers_journal_v0["cellpy_file_name"] = "cellpy_file_names"
headers_journal_v0["group"] = "groups"
headers_journal_v0["sub_group"] = "sub_groups"


def translate_headers(data_sets, cellpy_file_version):
    # this works for upgrading to versions 6,
    # remark that the extensions for the step table is hard-coded
    logging.debug(f"translate headers from v{cellpy_file_version}")

    summary_rename_dict = {
        headers_summary_v5[key]: HEADERS_SUMMARY[key]
        for key in HEADERS_SUMMARY
        if key in headers_summary_v5
    }

    steps_rename_dict = {
        headers_step_table_v5[key]: HEADERS_STEP_TABLE[key]
        for key in HEADERS_STEP_TABLE
        if key in headers_step_table_v5
    }

    steps_rename_dict_extensions = {}
    for key in HEADERS_KEYS_STEP_TABLE_EXTENDED:
        for extension in HEADERS_STEP_TABLE_EXTENSIONS:
            old = "_".join([HEADERS_STEP_TABLE[key], extension])
            new = "_".join([headers_step_table_v5[key], extension])
            steps_rename_dict_extensions[old] = new

    raw_rename_dict = {
        headers_normal_v5[key]: HEADERS_NORMAL[key] for key in HEADERS_NORMAL
    }

    summary_index_name = HEADERS_SUMMARY["cycle_index"]
    raw_index_name = HEADERS_NORMAL["data_point_txt"]

    # from pprint import pprint
    # pprint(summary_rename_dict)
    # pprint(steps_rename_dict)
    # pprint(steps_rename_dict_extensions)
    # pprint(raw_rename_dict)

    new_data_sets = []
    for data_set in data_sets:
        data_set.summary.rename(columns=summary_rename_dict, inplace=True)
        data_set.raw.rename(columns=raw_rename_dict, inplace=True)
        data_set.steps.rename(columns=steps_rename_dict, inplace=True)
        data_set.steps.rename(columns=steps_rename_dict_extensions, inplace=True)

        # we also need to update the index-name
        data_set.summary.index.name = summary_index_name
        data_set.raw.index.name = raw_index_name

        new_data_sets.append(data_set)

        # pprint(data_set.summary.columns)
        # pprint(data_set.steps.columns)
        # pprint(data_set.raw.columns)
    # check(new_data_sets)
    return new_data_sets


def check(data_sets):
    for data_set in data_sets:
        print(" checking ".center(80, "-"))
        print("RAW")
        print(data_set.raw.columns)
        print("SUMMARY")
        print(data_set.summary.columns)
        print("STEPS")
        print(data_set.steps.columns)
    print("OK?")



"""arbin MS SQL Server data"""
import datetime
import logging
import os
import platform
import shutil
import sys
import tempfile
import time
import warnings

import numpy as np
import pandas as pd
import pyodbc
from dateutil.parser import parse

from cellpy import prms
from cellpy.parameters.internal_settings import HeaderDict, get_headers_normal
from cellpy.readers.core import (
    Cell,
    FileID,
    check64bit,
    humanize_bytes,
    xldate_as_datetime,
)
from cellpy.readers.instruments.base import Loader

# TODO: @muhammad - get more meta data from the SQL db
# TODO: @jepe - update the batch functionality (including filefinder)
# TODO: @muhammad - make routine for "setting up the SQL Server" so that it is accessible and document it

DEBUG_MODE = prms.Reader.diagnostics  # not used
ALLOW_MULTI_TEST_FILE = prms._allow_multi_test_file  # not used
ODBC = prms._odbc
SEARCH_FOR_ODBC_DRIVERS = prms._search_for_odbc_driver  # not used
SQL_SERVER = prms.Instruments.Arbin["SQL_server"]
SQL_UID = prms.Instruments.Arbin["SQL_UID"]
SQL_PWD = prms.Instruments.Arbin["SQL_PWD"]
SQL_DRIVER = prms.Instruments.Arbin["SQL_Driver"]


# Names of the tables in the SQL Server db that is used by cellpy

# Not used anymore - maybe use a similar dict for the SQL table names (they are hard-coded at the moment)
TABLE_NAMES = {
    "normal": "Channel_Normal_Table",
    "global": "Global_Table",
    "statistic": "Channel_Statistic_Table",
    "aux_global": "Aux_Global_Data_Table",
    "aux": "Auxiliary_Table",
}

# Contains several headers not encountered yet in the Arbin SQL Server tables
summary_headers_renaming_dict = {
    "test_id_txt": "Test_ID",
    "data_point_txt": "Data_Point",
    "vmax_on_cycle_txt": "Vmax_On_Cycle",
    "charge_time_txt": "Charge_Time",
    "discharge_time_txt": "Discharge_Time",
}

# Contains several headers not encountered yet in the Arbin SQL Server tables
normal_headers_renaming_dict = {
    "aci_phase_angle_txt": "ACI_Phase_Angle",
    "ref_aci_phase_angle_txt": "Reference_ACI_Phase_Angle",
    "ac_impedance_txt": "AC_Impedance",
    "ref_ac_impedance_txt": "Reference_AC_Impedance",
    "charge_capacity_txt": "Charge_Capacity",
    "charge_energy_txt": "Charge_Energy",
    "current_txt": "Current",
    "cycle_index_txt": "Cycle_ID",
    "data_point_txt": "Data_Point",
    "datetime_txt": "Date_Time",
    "discharge_capacity_txt": "Discharge_Capacity",
    "discharge_energy_txt": "Discharge_Energy",
    "internal_resistance_txt": "Internal_Resistance",
    "is_fc_data_txt": "Is_FC_Data",
    "step_index_txt": "Step_ID",
    "sub_step_index_txt": "Sub_Step_Index",  # new
    "step_time_txt": "Step_Time",
    "sub_step_time_txt": "Sub_Step_Time",  # new
    "test_id_txt": "Test_ID",
    "test_time_txt": "Test_Time",
    "voltage_txt": "Voltage",
    "ref_voltage_txt": "Reference_Voltage",  # new
    "dv_dt_txt": "dV/dt",
    "frequency_txt": "Frequency",  # new
    "amplitude_txt": "Amplitude",  # new
    "channel_id_txt": "Channel_ID",  # new Arbin SQL Server
    "data_flag_txt": "Data_Flags",  # new Arbin SQL Server
    "test_name_txt": "Test_Name",  # new Arbin SQL Server
}


# Arbin SQL Server table headers (for both data df and stats df)
# --------------------------------------------------------------
# Test_ID
# Channel_ID
# Date_Time
# Data_Point
# Test_Time
# Step_Time
# Cycle_ID
# Step_ID
# Current
# Voltage
# Charge_Capacity
# Discharge_Capacity
# Charge_Energy
# Discharge_Energy
# Data_Flags
# Test_Name
# (all these headers are now implemented in the internal_settings


def from_arbin_to_datetime(n):
    # This is a hack that I have not verified yet
    # (what is so special with 1976.08.23? I have only encountered 1900.01.01 and 1970.01.01 before)
    # (or is the internal clock in our arbin machine set to 2014?)
    n /= 1_000_000_000_000
    temp = datetime.datetime(1976, 8, 23)
    delta = datetime.timedelta(days=n)
    time_object = temp + delta
    time_in_str = time_object.strftime("%y-%m-%d %H:%M:%S:%f")
    return time_in_str


class ArbinSQLLoader(Loader):
    """Class for loading arbin-data from MS SQL server."""

    name = "arbin_sql"

    def __init__(self, *args, **kwargs):
        """initiates the ArbinSQLLoader class"""
        self.arbin_headers_normal = (
            get_headers_normal()
        )  # the column headers defined by Arbin
        self.cellpy_headers_normal = (
            get_headers_normal()
        )  # the column headers defined by cellpy
        self.arbin_headers_global = self.get_headers_global()
        self.arbin_headers_aux_global = self.get_headers_aux_global()
        self.arbin_headers_aux = self.get_headers_aux()
        self.current_chunk = 0  # use this to set chunks to load
        self.server = SQL_SERVER

    @staticmethod
    def get_headers_normal():
        """Defines the so-called normal column headings for Arbin SQL Server"""
        # covered by cellpy at the moment
        return get_headers_normal()

    @staticmethod
    def get_headers_aux():
        """Defines the so-called auxiliary table column headings for Arbin SQL Server"""
        # not in use (yet)
        headers = HeaderDict()
        # - aux column headings (specific for Arbin)
        headers["test_id_txt"] = "Test_ID"
        headers["data_point_txt"] = "Data_Point"
        headers["aux_index_txt"] = "Auxiliary_Index"
        headers["data_type_txt"] = "Data_Type"
        headers["x_value_txt"] = "X"
        headers["x_dt_value"] = "dX_dt"
        return headers

    @staticmethod
    def get_headers_aux_global():
        """Defines the so-called auxiliary global column headings for Arbin SQL Server"""
        # not in use yet
        headers = HeaderDict()
        # - aux global column headings (specific for Arbin)
        headers["channel_index_txt"] = "Channel_Index"
        headers["aux_index_txt"] = "Auxiliary_Index"
        headers["data_type_txt"] = "Data_Type"
        headers["aux_name_txt"] = "Nickname"
        headers["aux_unit_txt"] = "Unit"
        return headers

    @staticmethod
    def get_headers_global():
        """Defines the so-called global column headings for Arbin SQL Server"""
        # not in use yet
        headers = HeaderDict()
        # - global column headings (specific for Arbin)
        headers["applications_path_txt"] = "Applications_Path"
        headers["channel_index_txt"] = "Channel_Index"
        headers["channel_number_txt"] = "Channel_Number"
        headers["channel_type_txt"] = "Channel_Type"
        headers["comments_txt"] = "Comments"
        headers["creator_txt"] = "Creator"
        headers["daq_index_txt"] = "DAQ_Index"
        headers["item_id_txt"] = "Item_ID"
        headers["log_aux_data_flag_txt"] = "Log_Aux_Data_Flag"
        headers["log_chanstat_data_flag_txt"] = "Log_ChanStat_Data_Flag"
        headers["log_event_data_flag_txt"] = "Log_Event_Data_Flag"
        headers["log_smart_battery_data_flag_txt"] = "Log_Smart_Battery_Data_Flag"
        headers["mapped_aux_conc_cnumber_txt"] = "Mapped_Aux_Conc_CNumber"
        headers["mapped_aux_di_cnumber_txt"] = "Mapped_Aux_DI_CNumber"
        headers["mapped_aux_do_cnumber_txt"] = "Mapped_Aux_DO_CNumber"
        headers["mapped_aux_flow_rate_cnumber_txt"] = "Mapped_Aux_Flow_Rate_CNumber"
        headers["mapped_aux_ph_number_txt"] = "Mapped_Aux_PH_Number"
        headers["mapped_aux_pressure_number_txt"] = "Mapped_Aux_Pressure_Number"
        headers["mapped_aux_temperature_number_txt"] = "Mapped_Aux_Temperature_Number"
        headers["mapped_aux_voltage_number_txt"] = "Mapped_Aux_Voltage_Number"
        headers[
            "schedule_file_name_txt"
        ] = "Schedule_File_Name"  # KEEP FOR CELLPY FILE FORMAT
        headers["start_datetime_txt"] = "Start_DateTime"
        headers["test_id_txt"] = "Test_ID"  # KEEP FOR CELLPY FILE FORMAT
        headers["test_name_txt"] = "Test_Name"  # KEEP FOR CELLPY FILE FORMAT
        return headers

    @staticmethod
    def get_raw_units():
        raw_units = dict()
        raw_units["current"] = 1.0  # A
        raw_units["charge"] = 1.0  # Ah
        raw_units["mass"] = 0.001  # g
        return raw_units

    @staticmethod
    def get_raw_limits():
        """returns a dictionary with resolution limits"""
        raw_limits = dict()
        raw_limits["current_hard"] = 0.000_000_000_000_1
        raw_limits["current_soft"] = 0.000_01
        raw_limits["stable_current_hard"] = 2.0
        raw_limits["stable_current_soft"] = 4.0
        raw_limits["stable_voltage_hard"] = 2.0
        raw_limits["stable_voltage_soft"] = 4.0
        raw_limits["stable_charge_hard"] = 0.001
        raw_limits["stable_charge_soft"] = 5.0
        raw_limits["ir_change"] = 0.00001
        return raw_limits

    # TODO: rename this (for all instruments) to e.g. load
    # TODO: implement more options (bad_cycles, ...)
    def loader(self, name, **kwargs):
        """returns a Cell object with loaded data.

        Loads data from arbin SQL server db.

        Args:
            name (str): name of the test

        Returns:
            new_tests (list of data objects)
        """
        new_tests = []

        data_df, stat_df = self._query_sql(name)
        aux_data_df = None  # Needs to be implemented
        meta_data = None  # Should be implemented

        # init data

        # selecting only one value (might implement multi-channel/id use later)
        test_id = data_df["Test_ID"].iloc[0]
        id_name = f"{SQL_SERVER}:{name}:{test_id}"

        channel_id = data_df["Channel_ID"].iloc[0]

        data = Cell()
        data.loaded_from = id_name
        data.channel_index = channel_id
        data.test_ID = test_id
        data.test_name = name

        # The following meta data is not implemented yet for SQL loader:
        data.channel_number = None
        data.creator = None
        data.item_ID = None
        data.schedule_file_name = None
        data.start_datetime = None

        # Generating a FileID project - needs to be updated to allow for db queries:
        fid = FileID(id_name)
        data.raw_data_files.append(fid)

        data.raw = data_df
        data.raw_data_files_length.append(len(data_df))
        data.summary = stat_df
        data = self._post_process(data)
        data = self.identify_last_data_point(data)
        new_tests.append(data)

        return new_tests

    def _post_process(self, data, **kwargs):
        # TODO: move this to parent

        fix_datetime = kwargs.pop("fix_datetime", True)
        set_index = kwargs.pop("set_index", True)
        rename_headers = kwargs.pop("rename_headers", True)
        extract_start_datetime = kwargs.pop("extract_start_datetime", True)

        # TODO:  insert post-processing and div tests here
        #    - check dtypes

        # Remark that we also set index during saving the file to hdf5 if
        #   it is not set.
        from pprint import pprint

        if rename_headers:
            columns = {}
            for key in self.arbin_headers_normal:
                old_header = normal_headers_renaming_dict.get(key, None)
                new_header = self.cellpy_headers_normal[key]
                if old_header:
                    columns[old_header] = new_header
                logging.debug(
                    f"processing cellpy normal header key '{key}':"
                    f" old_header='{old_header}' -> new_header='{new_header}'"
                )
            logging.debug(f"renaming dict: {columns}")
            data.raw.rename(index=str, columns=columns, inplace=True)
            try:
                columns = {}
                for key, old_header in summary_headers_renaming_dict.items():
                    try:
                        columns[old_header] = self.cellpy_headers_normal[key]
                    except KeyError:
                        columns[old_header] = old_header.lower()
                data.summary.rename(index=str, columns=columns, inplace=True)
            except Exception as e:
                logging.debug(f"Could not rename summary df ::\n{e}")

        if fix_datetime:

            h_datetime = self.cellpy_headers_normal.datetime_txt
            logging.debug("converting to datetime format")

            data.raw[h_datetime] = data.raw[h_datetime].apply(from_arbin_to_datetime)

            h_datetime = h_datetime
            if h_datetime in data.summary:
                data.summary[h_datetime] = data.summary[h_datetime].apply(
                    from_arbin_to_datetime
                )

        if set_index:
            hdr_data_point = self.cellpy_headers_normal.data_point_txt
            if data.raw.index.name != hdr_data_point:
                data.raw = data.raw.set_index(hdr_data_point, drop=False)

        if extract_start_datetime:
            hdr_date_time = self.arbin_headers_normal.datetime_txt
            data.start_datetime = parse("20" + data.raw[hdr_date_time].iat[0][:-7])

        return data

    def _query_sql(self, name):
        # TODO: refactor and include optional SQL arguments
        name_str = f"('{name}', '')"
        con_str = (
            f"Driver={{{SQL_DRIVER}}};"
            + f"Server={SQL_SERVER};UID={SQL_UID};PWD={SQL_PWD};Trusted_Connection=yes;"
        )

        # TODO: use variable for the name of the main db (ArbinPro8....)
        # TODO: consider making a function that searches for correct ArbinPro version
        master_q = (
            "SELECT Database_Name, Test_Name FROM "
            "ArbinPro8MasterInfo.dbo.TestList_Table WHERE "
            f"ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name IN {name_str}"
        )

        conn = pyodbc.connect(con_str)
        sql_query = pd.read_sql_query(master_q, conn)

        datas_df = []
        stats_df = []

        for index, row in sql_query.iterrows():
            # TODO: use variables - see above
            # TODO: consider to use f-strings
            data_query = (
                "SELECT "
                + str(row["Database_Name"])
                + ".dbo.IV_Basic_Table.*, ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name "
                "FROM " + str(row["Database_Name"]) + ".dbo.IV_Basic_Table "
                "JOIN ArbinPro8MasterInfo.dbo.TestList_Table "
                "ON "
                + str(row["Database_Name"])
                + ".dbo.IV_Basic_Table.Test_ID = ArbinPro8MasterInfo.dbo.TestList_Table.Test_ID "
                "WHERE ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name IN "
                + str(name_str)
            )

            stat_query = (
                "SELECT "
                + str(row["Database_Name"])
                + ".dbo.StatisticData_Table.*, ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name "
                "FROM " + str(row["Database_Name"]) + ".dbo.StatisticData_Table "
                "JOIN ArbinPro8MasterInfo.dbo.TestList_Table "
                "ON "
                + str(row["Database_Name"])
                + ".dbo.StatisticData_Table.Test_ID = ArbinPro8MasterInfo.dbo.TestList_Table.Test_ID "
                "WHERE ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name IN "
                + str(name_str)
            )

            datas_df.append(pd.read_sql_query(data_query, conn))
            stats_df.append(pd.read_sql_query(stat_query, conn))

        data_df = pd.concat(datas_df, axis=0)
        stat_df = pd.concat(stats_df, axis=0)

        return data_df, stat_df


def check_sql_loader(server: str = None, tests: list = None):
    test_name = tuple(tests) + ("",)  # neat trick :-)
    print(f"** test str: {test_name}")
    con_str = "Driver={SQL Server};Server=" + server + ";Trusted_Connection=yes;"
    master_q = (
        "SELECT Database_Name, Test_Name FROM "
        "ArbinPro8MasterInfo.dbo.TestList_Table WHERE "
        f"ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name IN {test_name}"
    )

    conn = pyodbc.connect(con_str)
    print("** connected to server")
    sql_query = pd.read_sql_query(master_q, conn)
    print("** SQL query:")
    print(sql_query)
    for index, row in sql_query.iterrows():
        # Muhammad, why is it a loop here?
        print(f"** index: {index}")
        print(f"** row: {row}")
        data_query = (
            "SELECT "
            + str(row["Database_Name"])
            + ".dbo.IV_Basic_Table.*, ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name "
            "FROM " + str(row["Database_Name"]) + ".dbo.IV_Basic_Table "
            "JOIN ArbinPro8MasterInfo.dbo.TestList_Table "
            "ON "
            + str(row["Database_Name"])
            + ".dbo.IV_Basic_Table.Test_ID = ArbinPro8MasterInfo.dbo.TestList_Table.Test_ID "
            "WHERE ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name IN "
            + str(test_name)
        )

        stat_query = (
            "SELECT "
            + str(row["Database_Name"])
            + ".dbo.StatisticData_Table.*, ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name "
            "FROM " + str(row["Database_Name"]) + ".dbo.StatisticData_Table "
            "JOIN ArbinPro8MasterInfo.dbo.TestList_Table "
            "ON "
            + str(row["Database_Name"])
            + ".dbo.StatisticData_Table.Test_ID = ArbinPro8MasterInfo.dbo.TestList_Table.Test_ID "
            "WHERE ArbinPro8MasterInfo.dbo.TestList_Table.Test_Name IN "
            + str(test_name)
        )
        print(f"** data query: {data_query}")
        print(f"** stat query: {stat_query}")

        # if looping, maybe these should be concatenated?
        data_df = pd.read_sql_query(data_query, conn)
        stat_df = pd.read_sql_query(stat_query, conn)

    return data_df, stat_df


def check_query():
    import pathlib

    name = ["20201106_HC03B1W_1_cc_01"]
    dd, ds = check_sql_loader(SQL_SERVER, name)
    out = pathlib.Path(r"C:\scripts\notebooks\Div")
    dd.to_clipboard()
    input("x")
    ds.to_clipboard()


def check_loader():
    print(" Testing connection to arbin sql server ".center(80, "-"))

    sql_loader = ArbinSQLLoader()
    name = "20201106_HC03B1W_1_cc_01"
    cell = sql_loader.loader(name)

    return cell


def check_loader_from_outside():
    import matplotlib.pyplot as plt

    from cellpy import cellreader

    name = "20200820_CoFBAT_slurry07B_01_cc_01"
    c = cellreader.CellpyData()
    c.set_instrument("arbin_sql")
    # print(c)
    c.from_raw(name)
    # print(c)
    c.make_step_table()
    c.make_summary()
    # print(c)
    raw = c.cell.raw
    steps = c.cell.steps
    summary = c.cell.summary
    raw.to_csv(r"C:\scripting\trash\raw.csv", sep=";")
    steps.to_csv(r"C:\scripting\trash\steps.csv", sep=";")
    summary.to_csv(r"C:\scripting\trash\summary.csv", sep=";")

    n = c.get_number_of_cycles()
    print(f"number of cycles: {n}")

    cycle = c.get_cap(1, method="forth")
    print(cycle.head())
    cycle.plot(x="capacity", y="voltage")
    plt.show()


def check_get():
    import cellpy

    name = "20200820_CoFBAT_slurry07B_01_cc_01"
    c = cellpy.get(name, instrument="arbin_sql")
    print(c)


if __name__ == "__main__":
    # test_query()
    # cell = test_loader()
    check_get()



# Works for data from KIT (SIMBA project)

file_info = {"raw_extension": "txt"}

unit_labels = {
    "resistance": "Ohms",
    "time": "s",
    "current": "A",
    "voltage": "V",
    "power": "W",
    "capacity": "Ah",
    "energy": "Wh",
    "temperature": "C",
}

normal_headers_renaming_dict = {
    "data_point_txt": "Rec",
    "cycle_index_txt": "Cycle C",
    "step_index_txt": "Step",
    "test_time_txt": "TestTime",
    "step_time_txt": "StepTime",
    "charge_capacity_txt": "Cap. [Ah]",
    "charge_energy_txt": "Ener. [Wh]",
    "current_txt": "Current [A]",
    "voltage_txt": "Voltage [V]",
    "datetime_txt": "DPT Time",
}

states = {
    "column_name": "Md",
    "charge_keys": ["C"],
    "discharge_keys": ["D"],
    "rest_keys": ["R"],
}


raw_units = {"current": 1.0, "charge": 1.0, "mass": 0.001}

raw_limits = {
    "current_hard": 0.000_000_000_000_1,
    "current_soft": 0.000_01,
    "stable_current_hard": 2.0,
    "stable_current_soft": 4.0,
    "stable_voltage_hard": 2.0,
    "stable_voltage_soft": 4.0,
    "stable_charge_hard": 0.001,
    "stable_charge_soft": 5.0,
    "ir_change": 0.00001,
}

formatters = {
    "skiprows": 12,  # 12 for other file
    "sep": "\t",
    "header": 0,  # 0 for other file
    "encoding": "ISO-8859-1",  # options: "ISO-8859-1", "utf-8", "cp1252"
    "decimal": ",",
    "thousands": None,
}

pre_processors = {
    "remove_empty_lines": True,
}

post_processors = {
    "split_capacity": True,
    "split_current": True,
    "set_index": True,
    "rename_headers": True,
    "remove_last_if_bad": True,
    "convert_date_time_to_datetime": True,
}



import json
import logging
import os
import pathlib
import platform
import shutil
import tempfile
import warnings

import pandas as pd

from cellpy.exceptions import UnderDefined
from cellpy.parameters import prms
from cellpy.parameters.internal_settings import (
    get_headers_journal,
    keys_journal_session,
)
from cellpy.parameters.legacy.internal_settings import (
    headers_journal_v0 as hdr_journal_old,
)
from cellpy.readers import dbreader
from cellpy.utils.batch_tools.batch_core import BaseJournal
from cellpy.utils.batch_tools.engines import simple_db_engine

hdr_journal = get_headers_journal()

trans_dict = {}
missing_keys = []
for key in hdr_journal:
    if key in hdr_journal_old:
        trans_dict[hdr_journal_old[key]] = hdr_journal[key]
    else:
        missing_keys.append(key)


class LabJournal(BaseJournal):
    def __init__(self, db_reader="default", engine=None):
        """Journal for selected batch.

        The journal contains pages (pandas.DataFrame) with prms for
        each cells (one cell pr row).

        Args:
            db_reader: either default (a simple excel reader already
                implemented in cellpy) or other db readers that implement
                the needed API.
            engine: defaults to simple_db_engine for parsing db using the
                db_reader
                    self.pages = simple_db_engine(
                        self.db_reader, id_keys, **kwargs
                    )
        """

        super().__init__()
        if db_reader == "default":
            self.db_reader = dbreader.Reader()
            self.engine = simple_db_engine
        else:
            logging.debug(f"Remark! db_reader: {db_reader}")
            self.db_reader = db_reader
            if engine is None:
                self.engine = engine
        self.batch_col = "b01"

    def _check_file_name(self, file_name, to_project_folder=False):
        if file_name is None:
            if not self.file_name:
                self.generate_file_name()
            file_name = pathlib.Path(self.file_name)
        else:
            file_name = pathlib.Path(file_name)

        if to_project_folder:
            file_name = file_name.with_suffix(".json")
            file_name = pathlib.Path(self.project_dir) / file_name
        self.file_name = file_name  # updates object (maybe not smart)
        return file_name

    def from_db(self, project=None, name=None, batch_col=None, **kwargs):
        logging.debug("creating journal from db")
        if batch_col is None:
            batch_col = self.batch_col
        if project is not None:
            self.project = project
        if name is None:
            name = self.name
        else:
            self.name = name
        logging.debug(f"batch_name, batch_col: {name}, {batch_col}")
        if self.db_reader is not None:
            id_keys = self.db_reader.select_batch(name, batch_col)
            self.pages = self.engine(self.db_reader, id_keys, **kwargs)
            if self.pages.empty:
                logging.critical(
                    f"EMPTY JOURNAL: are you sure you have provided correct input to batch?"
                )
                logging.critical(f"name: {name}")
                logging.critical(f"project: {self.project}")
                logging.critical(f"batch_col: {batch_col}")
        else:
            logging.debug("creating empty journal pages")
            self.pages = pd.DataFrame()

        self.generate_empty_session()
        self.generate_folder_names()
        self.paginate()

    def generate_empty_session(self):
        self.session = {}
        for item in keys_journal_session:
            self.session[item] = None

    @staticmethod
    def _fix_cellpy_paths(p):
        if platform.system() != "Windows":
            if p.find("\\") >= 0:
                # convert from win to posix
                p = pathlib.PureWindowsPath(p)
        else:
            if p.find("/") >= 0:
                # convert from posix to win
                p = pathlib.PurePosixPath(p)
        return pathlib.Path(p)

    @classmethod
    def read_journal_jason_file(cls, file_name, **kwargs):
        logging.debug(f"json loader starting on {file_name}")
        with open(file_name, "r") as infile:
            top_level_dict = json.load(infile)
        pages_dict = top_level_dict["info_df"]
        meta = top_level_dict["metadata"]
        session = top_level_dict.get("session", None)
        pages = pd.DataFrame(pages_dict)
        if pages.empty:
            logging.critical("could not find any pages in the journal")
            raise UnderDefined

        pages = cls._clean_pages(pages)

        if session is None:
            logging.debug(f"no session - generating empty one")
            session = dict()

        session, pages = cls._clean_session(session, pages)

        return pages, meta, session

    @classmethod
    def read_journal_excel_file(cls, file_name, **kwargs):
        sheet_names = {"meta": "meta", "pages": "pages", "session": "session"}
        project = kwargs.pop("project", "NaN")
        name = kwargs.pop("batch", pathlib.Path(file_name).stem)
        _meta = {
            "name": name,
            "project": project,
            "project_dir": pathlib.Path("."),
            "batch_dir": pathlib.Path("."),
            "raw_dir": pathlib.Path("."),
        }
        logging.debug(f"xlsx loader starting on {file_name}")

        meta_sheet_name = sheet_names["meta"]  # not tested yet
        pages_sheet_name = sheet_names["pages"]
        session_sheet_name = sheet_names["session"]  # not tested yet

        temporary_directory = tempfile.mkdtemp()
        temporary_file_name = shutil.copy(file_name, temporary_directory)
        try:
            pages = pd.read_excel(
                temporary_file_name, engine="openpyxl", sheet_name=pages_sheet_name
            )
        except KeyError:
            print(f"Worksheet '{pages_sheet_name}' does not exist.")
            return None

        try:
            session = pd.read_excel(
                temporary_file_name,
                sheet_name=session_sheet_name,
                engine="openpyxl",
                header=[0, 1],
            )
        except (KeyError, ValueError):
            print(f"Worksheet '{session_sheet_name}' does not exist.")
            session = None

        try:
            meta = pd.read_excel(
                temporary_file_name, sheet_name=meta_sheet_name, engine="openpyxl"
            )

        except (KeyError, ValueError):
            print(f"Worksheet '{meta_sheet_name}' does not exist.")
            meta = None

        if pages.empty:
            logging.critical("could not find any pages in the journal")
            raise UnderDefined
        pages = cls._clean_pages(pages)
        pages = pages.set_index(hdr_journal.filename)

        if meta is None:
            meta = _meta
        else:
            meta = cls._unpack_meta(meta) or _meta

        if session is None:
            logging.debug(f"no session - generating empty one")
            session = dict()
        else:
            session = cls._unpack_session(session)

        session, pages = cls._clean_session(session, pages)

        return pages, meta, session

    @classmethod
    def _unpack_session(cls, session):
        try:
            bcn2 = {
                l: list(sb["cycle_index"].values)
                for l, sb in session["bad_cycles"].groupby("cell_name")
            }
        except KeyError:
            bcn2 = []

        try:
            bc2 = list(session["bad_cells"]["cell_name"].dropna().values.flatten())
        except KeyError:
            bc2 = []

        try:
            s2 = list(session["starred"]["cell_name"].dropna().values.flatten())
        except KeyError:
            s2 = []

        try:
            n2 = list(session["notes"]["txt"].dropna().values.flatten())
        except KeyError:
            n2 = []

        session = {"bad_cycles": bcn2, "bad_cells": bc2, "starred": s2, "notes": n2}

        return session

    @classmethod
    def _unpack_meta(cls, meta):
        try:
            meta = meta.loc[:, ["parameter", "value"]]
        except KeyError:
            return
        meta = meta.set_index("parameter")
        return meta.to_dict()["value"]

    @classmethod
    def _clean_session(cls, session, pages):
        # include steps for cleaning up the session dict here
        if not session:
            logging.critical("no session found in your journal file")
        for item in keys_journal_session:
            session[item] = session.get(item, None)

        return session, pages

    @classmethod
    def _clean_pages(cls, pages):
        import ast

        logging.debug("removing empty rows")
        pages = pages.dropna(how="all")
        logging.debug("checking path-names")
        try:
            p = pages[hdr_journal.raw_file_names]
            new_p = []
            for f in p:
                if isinstance(f, str):
                    try:
                        new_f = ast.literal_eval(f)
                        if isinstance(new_f, list):
                            f = new_f
                    except Exception as e:
                        warnings.warn(e)
                        warnings.warn(f"Could not evaluate {f}")

                new_p.append(f)
            pages[hdr_journal.raw_file_names] = new_p

        except KeyError:
            print(
                "Tried but failed in converting raw_file_names into an appropriate list"
            )
        try:
            pages[hdr_journal.cellpy_file_name] = pages[
                hdr_journal.cellpy_file_name
            ].apply(cls._fix_cellpy_paths)
        except KeyError:
            # assumes it is an old type journal file
            print(f"The key '{hdr_journal.cellpy_file_name}' is missing!")
            print(f"Assumes that this is an old-type journal file.")
            try:
                pages.rename(columns=trans_dict, inplace=True)
                pages[hdr_journal.cellpy_file_name] = pages[
                    hdr_journal.cellpy_file_name
                ].apply(cls._fix_cellpy_paths)
                logging.warning("old journal file - updating")
            except KeyError:
                print("Error! Could still not parse the pages.")
                print(f"Missing key: {hdr_journal.cellpy_file_name}")
                pages[hdr_journal.cellpy_file_name] = None

        # only keep selected cells if keep column exists
        if "keep" in pages.columns:
            logging.debug("Journal contains 'keep' - selecting only 'keep' > 0.")
            pages = pages.loc[pages.keep > 0, :]

        for column_name in missing_keys:
            if column_name not in pages.columns:
                logging.debug(f"wrong journal format - missing: {column_name}")
                pages[column_name] = None

        for column_name in hdr_journal:
            if column_name not in pages.columns:
                if column_name != hdr_journal.filename:
                    pages[column_name] = None

        return pages

    def from_file(self, file_name=None, paginate=True, **kwargs):
        """Loads a DataFrame with all the needed info about the experiment"""

        file_name = self._check_file_name(file_name)
        logging.debug(f"reading {file_name}")
        if pathlib.Path(file_name).suffix.lower() == ".xlsx":
            file_loader = self.read_journal_excel_file
        else:
            file_loader = self.read_journal_jason_file
        try:
            out = file_loader(file_name, **kwargs)
            if out is None:
                raise IOError(f"Error reading {file_name}.")
            pages, meta_dict, session = out
        except UnderDefined as e:
            logging.critical(f"could not load {file_name}")
            raise UnderDefined from e

        logging.debug(f"got pages and meta_dict")

        self.pages = pages
        self.session = session
        self.file_name = file_name
        self._prm_packer(meta_dict)

        if paginate:
            self.generate_folder_names()
            self.paginate()

    def from_frame(self, frame, name=None, project=None, paginate=None, **kwargs):
        if name is not None:
            self.name = name
        if project is not None:
            self.project = project

        self.pages = (
            frame  # TODO: include a check here to see if the pages are appropriate
        )

        if paginate is None:
            if self.name and self.project:
                paginate = True

        if paginate:
            logging.critical(f"paginating {project}/{name} ")
            self.generate_folder_names()
            self.paginate()

    def from_file_old(self, file_name=None):
        """Loads a DataFrame with all the needed info about the experiment"""

        file_name = self._check_file_name(file_name)

        with open(file_name, "r") as infile:
            top_level_dict = json.load(infile)

        pages_dict = top_level_dict["info_df"]
        pages = pd.DataFrame(pages_dict)
        pages[hdr_journal.cellpy_file_name] = pages[hdr_journal.cellpy_file_name].apply(
            self._fix_cellpy_paths
        )
        self.pages = pages
        self.file_name = file_name
        self._prm_packer(top_level_dict["metadata"])
        self.generate_folder_names()
        self.paginate()

    def create_empty_pages(self, description=None):
        if description is not None:
            print(f"Creating from {type(description)} is not implemented yet")

        logging.debug("Creating an empty journal")
        logging.debug(f"name: {self.name}")
        logging.debug(f"project: {self.project}")

        col_names = list(hdr_journal.values())
        pages = pd.DataFrame(columns=col_names)
        pages.set_index(hdr_journal.filename, inplace=True)
        return pages

    def to_file(self, file_name=None, paginate=True, to_project_folder=True):
        """Saves a DataFrame with all the needed info about the experiment.

        Args:
            file_name (str or pathlib.Path): journal file name
            paginate (bool): make project folders
            to_project_folder (bool): save journal file to the folder containing your cellpy projects

        Returns:

        """
        file_name = self._check_file_name(
            file_name, to_project_folder=to_project_folder
        )
        pages = self.pages
        session = self.session
        meta = self._prm_packer()
        top_level_dict = {"info_df": pages, "metadata": meta, "session": session}

        is_json = False
        is_xlsx = False

        if file_name.suffix == ".xlsx":
            is_xlsx = True

        if file_name.suffix == ".json":
            is_json = True

        if is_xlsx:
            df_session = self._pack_session(session)
            df_meta = self._pack_meta(meta)

            try:
                with pd.ExcelWriter(file_name, mode="w", engine="openpyxl") as writer:
                    pages.to_excel(writer, sheet_name="pages", engine="openpyxl")
                    # no index is not supported for multi-index (update to index=False when pandas implement it):
                    df_session.to_excel(writer, sheet_name="session", engine="openpyxl")
                    df_meta.to_excel(
                        writer, sheet_name="meta", engine="openpyxl", index=False
                    )
            except PermissionError as e:
                print(f"Could not load journal to xlsx ({e})")

        if is_json:
            jason_string = json.dumps(
                top_level_dict,
                default=lambda info_df: json.loads(
                    info_df.to_json(default_handler=str)
                ),
            )

            with open(file_name, "w") as outfile:
                outfile.write(jason_string)

        self.file_name = file_name
        logging.info(f"Saved file to {file_name}")

        if paginate:
            self.paginate()

    @staticmethod
    def _pack_session(session):
        frames = []
        keys = []
        try:
            l_bad_cycle_numbers = []

            for k, v in session["bad_cycles"].items():
                l_bad_cycle_numbers.append(pd.DataFrame(data=v, columns=[k]))

            df_bad_cycle_numbers = (
                pd.concat(l_bad_cycle_numbers, axis=1)
                .melt(var_name="cell_name", value_name="cycle_index")
                .dropna()
            )
            frames.append(df_bad_cycle_numbers)
            keys.append("bad_cycles")
        except (KeyError, AttributeError):
            logging.debug("missing bad cycle numbers")

        df_bad_cells = pd.DataFrame(session["bad_cells"], columns=["cell_name"])
        frames.append(df_bad_cells)
        keys.append("bad_cells")

        df_starred = pd.DataFrame(session["starred"], columns=["cell_name"])
        frames.append(df_starred)
        keys.append("starred")

        df_notes = pd.DataFrame(session["notes"], columns=["txt"])
        frames.append(df_notes)
        keys.append("notes")

        session = pd.concat(frames, axis=1, keys=keys)
        return session

    @staticmethod
    def _pack_meta(meta):
        meta = pd.DataFrame(meta, index=[0]).melt(
            var_name="parameter", value_name="value"
        )
        return meta

    def generate_folder_names(self):
        """Set appropriate folder names."""
        logging.debug("creating folder names")
        if self.project and isinstance(self.project, (pathlib.Path, str)):
            logging.debug("got project name")
            logging.debug(self.project)
            self.project_dir = os.path.join(prms.Paths.outdatadir, self.project)
        else:
            logging.critical(
                "Could not create project dir (missing project definition)"
            )
        if self.name:
            self.batch_dir = os.path.join(self.project_dir, self.name)
            self.raw_dir = os.path.join(self.batch_dir, "raw_data")
        else:
            logging.critical(
                "Could not create batch_dir and raw_dir", "(missing batch name)"
            )
        logging.debug(f"batch dir: {self.batch_dir}")
        logging.debug(f"project dir: {self.project_dir}")
        logging.debug(f"raw dir: {self.raw_dir}")

    def paginate(self):
        """Make folders where we would like to put results etc."""

        project_dir = self.project_dir
        raw_dir = self.raw_dir
        batch_dir = self.batch_dir

        if project_dir is None:
            raise UnderDefined("no project directory defined")
        if raw_dir is None:
            raise UnderDefined("no raw directory defined")
        if batch_dir is None:
            raise UnderDefined("no batch directory defined")

        # create the folders
        if not os.path.isdir(project_dir):
            os.mkdir(project_dir)
            logging.info(f"created folder {project_dir}")
        if not os.path.isdir(batch_dir):
            os.mkdir(batch_dir)
            logging.info(f"created folder {batch_dir}")
        if not os.path.isdir(raw_dir):
            os.mkdir(raw_dir)
            logging.info(f"created folder {raw_dir}")

        self.project_dir = project_dir
        self.batch_dir = batch_dir
        self.raw_dir = raw_dir

        return project_dir, batch_dir, raw_dir

    def generate_file_name(self):
        """generate a suitable file name for the experiment"""
        if not self.project:
            raise UnderDefined("project name not given")

        out_data_dir = prms.Paths.outdatadir
        project_dir = os.path.join(out_data_dir, self.project)
        file_name = f"cellpy_batch_{self.name}.json"
        self.file_name = os.path.join(project_dir, file_name)

    # v.1.0.0:
    def look_for_file(self):
        pass

    def get_column(self, header):
        """populate new column from db"""
        pass

    def get_cell(self, id_key):
        """get additional cell info from db"""
        pass

    def add_comment(self, comment):
        """add a comment (will be saved in the journal file)"""
        pass

    def remove_comment(self, comment_id):
        pass

    def view_comments(self):
        pass

    def remove_cell(self, cell_id):
        pass

    def add_cell(self, cell_id, **kwargs):
        """Add a cell to the pages"""
        pass


def _dev_journal_loading():
    from cellpy import log

    log.setup_logging(default_level="DEBUG")
    journal_file = pathlib.Path("../../../dev_data/db/test_journal.xlsx")
    print(journal_file.is_file())
    logging.debug(f"reading journal file {journal_file}")
    journal = LabJournal(db_reader=None)
    journal.from_file(journal_file, paginate=False)
    print(80 * "-")
    print(journal.pages)
    print(80 * "-")
    print(journal.session)

    # creating a mock session
    bad_cycle_numbers = {
        "20160805_test001_45_cc": [4, 337, 338],
        "20160805_test001_47_cc": [7, 8, 9],
    }
    bad_cells = ["20160805_test001_45_cc"]

    notes = {"date_stamp": "one comment for the road", "date_stamp2": "another comment"}
    session = {
        "bad_cycle_numbers": bad_cycle_numbers,
        "bad_cells": bad_cells,
        "notes": notes,
    }

    # journal.session = session

    new_journal_name = journal_file.with_name(f"{journal_file.stem}_tmp.xlsx")
    print(new_journal_name)
    journal.to_file(file_name=new_journal_name, paginate=False, to_project_folder=False)


if __name__ == "__main__":
    print(" running journal ".center(80, "-"))
    _dev_journal_loading()
    print(" finished ".center(80, "-"))
